from enum import Enum, auto

__all__ = ["LrcType", "LrcMetaType", "LinkType"]


class LrcType(Enum):
    Origin = auto()
    Translation = auto()
    Romaji = auto()

    def prettyString(self) -> str:
        match self:
            case LrcType.Origin:
                return "源"
            case LrcType.Translation:
                return "译"
            case LrcType.Romaji:
                return "音"


class LrcMetaType(Enum):
    Title = "ti"
    Artist = "ar"
    Album = "al"
    Author = "au"
    Length = "length"
    LrcAuthor = "by"
    Offset = "offset"


class LinkType(Enum):
    Track = "track"
    Album = "album"
    Playlist = "playlist"



from pathlib import Path

from click import Path as clickPath
from click import argument, command, option

from .app import NCMLyricsApp


@command
@option(
    "-o",
    "--outputs",
    type=clickPath(exists=True, file_okay=False, dir_okay=True, writable=True, path_type=Path),
    multiple=True,
    help="输出目录，输出文件名将自动匹配到已经存在的音频文件，重复指定此参数多次以实现回落匹配。",
)
@option("-e", "--exist", is_flag=True, help="仅在源文件存在时保存歌词文件。")
@option("-O", "--overwrite", is_flag=True, help="在歌词文件已存在时重新获取歌词并覆盖写入。")
@option("-q", "--quiet", is_flag=True, help="不进行任何提示并跳过所有确认。")
@argument(
    "links",
    nargs=-1,
)
def main(exist: bool, overwrite: bool, quiet: bool, outputs: list[Path], links: list[str]) -> None:
    NCMLyricsApp(exist=exist, overwrite=overwrite, quiet=quiet, outputs=outputs, links=links).run()


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from platform import system
from re import compile as compileRegex
from urllib.parse import parse_qs as parseQuery
from urllib.parse import urlparse as parseUrl

from httpx import get as getHttp

from .enum import LinkType
from .error import ParseLinkError, UnsupportedLinkError

__all__ = ["Link", "parseLink", "safeFileName"]

RE_SHARE_LINK_ID_BY_PATH = compileRegex(r"^/?(?P<id>\d+)$")
RE_SHARE_LINK_ANDROID_ALBUM_PATH = compileRegex(r"^/album/(?P<id>\d+)/?$")

if system() == "Windows":
    TRANSLATER_SAFE_FILENAME = str.maketrans(
        {i: 0x5F for i in (0x2F, 0x5C, 0x3A, 0x2A, 0x3F, 0x22, 0x3C, 0x3E, 0x7C)}
    )  # /, \, :, *, ?, ", <, >, | => _
else:
    TRANSLATER_SAFE_FILENAME = str.maketrans({0x2F: 0x5F})  # / => _


@dataclass
class Link:
    type: LinkType
    id: int


def parseLink(url: str) -> Link:
    parsedUrl = parseUrl(url, allow_fragments=False)
    contentType: LinkType | None = None
    contentId: int | None = None

    match parsedUrl.scheme:
        case "http" | "https":
            match parsedUrl.netloc:
                case "music.163.com":
                    match parsedUrl.path:
                        case "/playlist" | "/#/playlist":
                            contentType = LinkType.Playlist
                        case "/album" | "/#/album":
                            contentType = LinkType.Album
                        case "/song" | "/#/song":
                            contentType = LinkType.Track
                        case _:
                            # Hack for android client shared album link
                            matchedPath = RE_SHARE_LINK_ANDROID_ALBUM_PATH.match(parsedUrl.path)
                            if matchedPath is not None:
                                contentType = LinkType.Album
                                contentId = int(matchedPath["id"])
                            else:
                                raise UnsupportedLinkError(parsedUrl)
                case "y.music.163.com":
                    match parsedUrl.path:
                        case "/m/playlist":
                            contentType = LinkType.Playlist
                        case "/m/song":
                            contentType = LinkType.Track
                        case _:
                            raise UnsupportedLinkError(parsedUrl)
                case "163cn.tv":
                    response = getHttp(url)
                    if response.status_code != 302:
                        raise ParseLinkError(f"未知的 Api 响应: {response.status_code}")
                    newUrl = response.headers.get("Location")
                    if newUrl is None:
                        raise ParseLinkError("Api 未返回重定向结果")
                    return parseLink(newUrl)
                case _:
                    raise UnsupportedLinkError(parsedUrl)
        case "ncmlyrics":  # eg: ncmlyrics://playlist/123456, ncmlyrics://album/12456, ncmlyrics://track/123456
            try:
                contentType = LinkType(parsedUrl.netloc)
            except ValueError:
                raise UnsupportedLinkError(parsedUrl)

            if parsedUrl.path:
                matched = RE_SHARE_LINK_ID_BY_PATH.match(parsedUrl.path)
                if matched is not None:
                    contentId = int(matched.group("id"))
            else:
                raise ParseLinkError
        case "playlist" | "album" | "track":  # eg: playlist:123456, album:/12456, track://123456
            try:
                contentType = LinkType(parsedUrl.scheme)
            except ValueError:
                raise UnsupportedLinkError(parsedUrl)

            try:
                if parsedUrl.netloc:
                    contentId = int(parsedUrl.netloc)
                elif parsedUrl.path:
                    matched = RE_SHARE_LINK_ID_BY_PATH.match(parsedUrl.path)
                    if matched is not None:
                        contentId = int(matched.group("id"))
                else:
                    raise ParseLinkError
            except ValueError:
                raise ParseLinkError
        case _:
            raise UnsupportedLinkError(parsedUrl)

    if contentId is None:
        try:
            contentId = int(parseQuery(parsedUrl.query).get("id")[0])
        except Exception:
            raise ParseLinkError

    return Link(contentType, contentId)


def safeFileName(filename: str) -> str:
    return filename.translate(TRANSLATER_SAFE_FILENAME)



from pathlib import Path
from unittest import TestCase

from ncmlyrics.enum import LinkType
from ncmlyrics.error import ParseLinkError, UnsupportedLinkError
from ncmlyrics.object import NCMTrack
from ncmlyrics.util import Link, parseLink, pickOutput, testExistTrackSource


class TestUtils(TestCase):
    def test_parseLink_Windows(self):
        self.assertEqual(
            parseLink("https://music.163.com/playlist?id=444817519"),
            Link(LinkType.Playlist, 444817519),
            msg="Shared playlist from NCM Windows Client",
        )

        self.assertEqual(
            parseLink("https://music.163.com/album?id=34609577"),
            Link(LinkType.Album, 34609577),
            msg="Shared album from NCM Windows Client",
        )

        self.assertEqual(
            parseLink("https://music.163.com/song?id=2621105420"),
            Link(LinkType.Track, 2621105420),
            msg="Shared song from NCM Windows Client",
        )

    def test_parseLink_Website(self):
        self.assertEqual(
            parseLink("https://music.163.com/#/playlist?id=444817519"),
            Link(LinkType.Playlist, 444817519),
            msg="Playlist from NCM Website",
        )

        self.assertEqual(
            parseLink("https://music.163.com/#/album?id=34609577"),
            Link(LinkType.Album, 34609577),
            msg="Album from NCM Website",
        )

        self.assertEqual(
            parseLink("https://music.163.com/#/song?id=2621105420"),
            Link(LinkType.Track, 2621105420),
            msg="Song from NCM Website",
        )

    def test_parseLink_Android(self):
        self.assertEqual(
            parseLink("https://y.music.163.com/m/playlist?id=2224276126"),
            Link(LinkType.Playlist, 2224276126),
            msg="Shared playlist from NCM Android Client",
        )

        self.assertEqual(
            parseLink("http://music.163.com/album/3139945/"),
            Link(LinkType.Album, 3139945),
            msg="Shared album from NCM Android Client",
        )

        self.assertEqual(
            parseLink("https://y.music.163.com/m/song?id=2604307454"),
            Link(LinkType.Track, 2604307454),
            msg="Shared song from NCM Android Client",
        )

    def test_parseLink_302(self):
        self.assertEqual(
            parseLink("http://163cn.tv/xpaQwii"),
            Link(LinkType.Track, 413077069),
            msg="Shared song from NCM Android Client player",
        )

    def test_parseLink_special(self):
        self.assertEqual(
            parseLink("ncmlyrics://playlist/123456"),
            Link(LinkType.Playlist, 123456),
        )

        self.assertEqual(
            parseLink("ncmlyrics://album/123456"),
            Link(LinkType.Album, 123456),
        )

        self.assertEqual(
            parseLink("ncmlyrics://track/123456"),
            Link(LinkType.Track, 123456),
        )

        self.assertEqual(
            parseLink("playlist:123456"),
            Link(LinkType.Playlist, 123456),
        )

        self.assertEqual(
            parseLink("album:/123456"),
            Link(LinkType.Album, 123456),
        )

        self.assertEqual(
            parseLink("track://123456"),
            Link(LinkType.Track, 123456),
        )

    def test_parseLink_UnsupportedLinkError(self):
        self.assertRaises(
            UnsupportedLinkError,
            parseLink,
            "ftp://ftpserver.com/",
        )

        self.assertRaises(
            UnsupportedLinkError,
            parseLink,
            "https://www.google.com/",
        )

        self.assertRaises(
            UnsupportedLinkError,
            parseLink,
            "https://music.163.com/unsupport?id=123",
        )

        self.assertRaises(
            UnsupportedLinkError,
            parseLink,
            "https://music.163.com/album/123a",
        )

        self.assertRaises(
            UnsupportedLinkError,
            parseLink,
            "ncmlyrics://unsupport/123456",
        )

    def test_parseLink_ParseLinkError(self):
        self.assertRaises(
            ParseLinkError,
            parseLink,
            "https://music.163.com/playlist?id=123a",
        )

        self.assertRaises(
            ParseLinkError,
            parseLink,
            "playlist://123456a",
        )

    def test_testExistTrackSource(self):
        resources = Path("tests/resource/util/testExistTrackSource")

        self.assertEqual(
            testExistTrackSource(NCMTrack(0, "Mp3Name", ["Mp3Artist"]), resources),
            resources / "Mp3Artist - Mp3Name.mp3",
        )

        self.assertEqual(
            testExistTrackSource(NCMTrack(0, "Mp3Name", ["Mp3Artist1", "Mp3Artist2"]), resources),
            resources / "Mp3Artist1,Mp3Artist2 - Mp3Name.mp3",
        )

        self.assertEqual(
            testExistTrackSource(NCMTrack(0, "FlacName", ["FlacArtist1", "FlacArtist2"]), resources),
            resources / "FlacArtist1 FlacArtist2 - FlacName.flac",
        )

        self.assertEqual(
            testExistTrackSource(NCMTrack(0, "NcmNameWith.", ["NcmArtistWith."]), resources),
            resources / "NcmArtistWith. - NcmNameWith..ncm",
        )

    def test_pickOutput(self):
        resources = Path("tests/resource/util/pickOutput")

        self.assertEqual(
            pickOutput(NCMTrack(0, "testTrack", ["testArtist"]), []),
            Path("testArtist - testTrack.lrc"),
        )

        self.assertEqual(
            pickOutput(NCMTrack(0, "testTrack", ["testArtist"]), [resources / "1"]),
            resources / "1" / "testArtist - testTrack.lrc",
        )

        self.assertEqual(
            pickOutput(NCMTrack(0, "testTrack1", ["testArtist"]), [resources / "1", resources / "2"]),
            resources / "1" / "testArtist - testTrack1.lrc",
        )

        self.assertEqual(
            pickOutput(NCMTrack(0, "testTrack2", ["testArtist"]), [resources / "1", resources / "2"]),
            resources / "2" / "testArtist - testTrack2.lrc",
        )

        self.assertEqual(
            pickOutput(NCMTrack(0, "testTrack1", ["testArtist"]), [resources / "2", resources / "1"]),
            resources / "1" / "testArtist - testTrack1.lrc",
        )

        self.assertEqual(
            pickOutput(NCMTrack(0, str(), [str()]), [], True),
            None,
        )

        self.assertEqual(
            pickOutput(NCMTrack(0, "testTrack0", ["testArtist"]), [resources / "1"], True),
            None,
        )



from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import wait as waitFuture
from pathlib import Path
from re import Pattern
from re import compile as compileRegex
from re import escape as escapeRegex
from typing import Any, Generator, Iterable

from click import confirm
from rich.console import Console
from rich.progress import Progress, TaskID
from rich.theme import Theme

from .api import NCMApi
from .enum import LinkType
from .error import ParseLinkError, UnsupportedLinkError
from .lrc import Lrc
from .object import NCMAlbum, NCMPlaylist, NCMTrack
from .util import parseLink, safeFileName

__all__ = ["NCMLyricsApp"]

NCMLyricsAppTheme = Theme(
    {
        "tracktitle": "bold chartreuse1",
        "trackarrow": "chartreuse3",
        "albumtitle": "bold orchid1",
        "albumarrow": "orchid2",
        "playlisttitle": "bold aquamarine1",
        "playlistarrow": "aquamarine3",
        "info": "",
        "warning": "orange1",
        "error": "bold red1",
    }
)


class NCMLyricsApp:
    def __init__(self, exist: bool, overwrite: bool, quiet: bool, outputs: list[Path], links: list[str]) -> None:
        self.console = Console(theme=NCMLyricsAppTheme, highlight=False)
        self.progress = Progress(console=self.console)
        self.pool = ThreadPoolExecutor(max_workers=4)

        self.api = NCMApi()

        self.exist = exist
        self.overwrite = overwrite
        self.quiet = quiet
        if len(outputs) == 0:
            self.outputs = [Path()]
        else:
            self.outputs = outputs

        self.links = links

        self.tasks: list[NCMTrack | NCMAlbum | NCMPlaylist] = []
        self.tracks: list[NCMTrack] = []
        self.trackPairs: list[tuple[NCMTrack, Path | None]] = []

        self.existedTargets: list[Path] = []

    def run(self) -> None:
        if len(self.links) == 0:
            self.console.print(
                "请给出至少一个链接以解析曲目以获取其歌词！支持输入单曲，专辑与歌单的分享或网页链接。", style="error"
            )
            return

        progressId = self.progress.add_task("解析链接", total=len(self.links))
        self.progress.start()

        for task in self.pool.map(self.resolveLink, self._repeat(progressId), self.links):
            self.tasks.append(task)
            self.tracks.extend(task.tracks)

        if not self.quiet:
            self.progress.stop()
            self.printTasks()
            if not confirm("继续操作？", default=True):
                self.console.print("任务已取消。", style="info")
                return

        self.progress.reset(progressId, description="获取已存在的歌曲列表", total=1)

        if not self.quiet:
            self.progress.start()
        self.walkOutputs()
        self.progress.advance(progressId)

        self.progress.reset(progressId, description="解析保存路径", total=len(self.tracks))

        for trackPair in self.pool.map(self.resolvePath, self._repeat(progressId), self.tracks):
            self.trackPairs.append(trackPair)

        self.progress.reset(progressId, description="输出 Lrc 文件", total=len(self.trackPairs))

        waitFuture((self.pool.submit(self.exportLrc, progressId, *trackPair) for trackPair in self.trackPairs))

        self.progress.stop()
        self.api.saveCookies()

    def printTasks(self):
        def printTracks(tracks: Iterable[NCMTrack], arrowStyle: str | None = None) -> None:
            for track in tracks:
                self.console.print(
                    f"[{arrowStyle}]-->[/{arrowStyle}] [link={track.link()}]{track.prettyString()}[/link]"
                )

        for task in self.tasks:
            match task:
                case NCMTrack():
                    self.console.print(
                        f"[tracktitle]-- 单曲 -->[/tracktitle] [link={task.link()}]{task.prettyString()}[/link]"
                    )
                case NCMAlbum():
                    self.console.print(f"[albumtitle]== 专辑 ==>[/albumtitle] [link={task.link()}]{task.name}[/link]")
                    printTracks(task.tracks, "albumarrow")
                case NCMPlaylist():
                    self.console.print(
                        f"[playlisttitle]== 歌单 ==>[/playlisttitle] [link={task.link()}]{task.name}[/link]"
                    )
                    printTracks(task.tracks, "playlistarrow")

    def walkOutputs(self):
        for output in self.outputs:
            output = output.absolute()
            if not output.exists() or not output.is_dir():
                continue
            for content in output.iterdir():
                if not content.is_file():
                    continue
                if content.suffix in (".ncm", ".mp3", ".flac"):
                    self.existedTargets.append(content)

    def resolveLink(self, progress: TaskID, link: str) -> NCMTrack | NCMAlbum | NCMPlaylist:
        try:
            parsed = parseLink(link)
        except UnsupportedLinkError:
            self.console.print(f"不支持的链接：{link}", style="error")
            return
        except ParseLinkError:
            self.console.print_exception()
            self.console.print(f"解析链接时出现错误：{link}", style="error")
            return

        match parsed.type:
            case LinkType.Track:
                result = self.api.getDetailsForTrack(parsed.id)
            case LinkType.Album:
                result = self.api.getDetailsForAlbum(parsed.id)
            case LinkType.Playlist:
                result = self.api.getDetailsForPlaylist(parsed.id)
                result.fillDetailsOfTracks(self.api)

        self.progress.advance(progress)
        return result

    def resolvePath(self, progress: TaskID, track: NCMTrack) -> tuple[NCMTrack, Path | None]:
        regex: Pattern[str] | None = None
        targetLrc: Path | None = None

        for target in self.existedTargets:
            if regex is None:
                # "(?:," + ")?(?:,".join((escapeRegex(artist) for artist in track.artists[3:])) + ")?"
                regex = compileRegex(
                    rf"^{escapeRegex(",".join(track.artists[:3])).replace(",", "(?:,| )")} - {escapeRegex(track.name.rstrip("."))}.+(ncm|mp3|flac)$"
                )
            matched = regex.match(target.name)
            if matched is not None:
                targetLrc = target.with_suffix(".lrc")
                break

        self.progress.advance(progress)

        if targetLrc is None:
            if self.exist:
                return (track, None)
            else:
                targetLrc = self.outputs[-1] / safeFileName(f"{",".join(track.artists)} - {track.name}.lrc")

        if not self.overwrite and targetLrc.exists():
            return (track, None)
        return (track, targetLrc)

    def exportLrc(self, progress: TaskID, track: NCMTrack, path: Path | None) -> None:
        if path is None:
            self.console.print(
                f"[trackarrow]-->[/trackarrow] {track.prettyString()} [dark_turquoise]==>[dark_turquoise] [warning]对应的歌词文件已存在, 跳过此曲目。[/warning]"
            )
            self.progress.advance(progress)
            return

        ncmlyrics = self.api.getLyricsByTrack(track.id)
        if ncmlyrics.isPureMusic:
            self.console.print(
                f"[trackarrow]-->[/trackarrow] {track.prettyString()} [dark_turquoise]==>[/dark_turquoise] [warning]为纯音乐, 跳过此曲目。[/warning]"
            )
        else:
            if not self.quiet:
                self.console.print(
                    f"[trackarrow]-->[/trackarrow] {track.prettyString()} [dark_turquoise]==>[/dark_turquoise] [info]{str(path)}[/info]"
                )
            Lrc.fromNCMLyrics(ncmlyrics).saveAs(path)

        self.progress.advance(progress)

    @staticmethod
    def _repeat(content: Any) -> Generator[Any, None, None]:
        while True:
            yield content



from httpx import URL
from platformdirs import PlatformDirs

from .__version__ import __title__

NCM_API_BASE_URL = URL("https://interface.music.163.com/api")

CONFIG_LRC_AUTO_MERGE = True
CONFIG_LRC_AUTO_MERGE_OFFSET = 50

CONFIG_API_DETAIL_TRACK_PER_REQUEST = 150

PLATFORM = PlatformDirs(appname=__title__, ensure_exists=True)



from dataclasses import dataclass
from json import JSONDecodeError
from typing import Self

from httpx import Response

from .enum import LrcType
from .error import ObjectParseError

__all__ = ["NCMTrack", "NCMAlbum", "NCMPlaylist", "NCMLyrics"]


@dataclass
class NCMTrack:
    id: int
    name: str
    artists: list[str]

    @classmethod
    def fromApi(cls, response: Response) -> list[Self]:
        try:
            data: dict = response.json()
        except JSONDecodeError:
            raise ObjectParseError("无法以预期的 Json 格式解析响应")

        if data.get("code") != 200:
            raise ObjectParseError(f"响应码不为 200: {data["code"]}")

        data = data.get("songs")
        if data is None:
            raise ObjectParseError("不存在单曲对应的结构", data)

        result = []

        for track in data:
            result.append(cls.fromData(track))

        return result

    @classmethod
    def fromData(cls, data: dict) -> Self:
        try:
            return cls(
                id=data["id"],
                name=data["name"],
                artists=[artist["name"] for artist in data["ar"]],
            )
        except KeyError as e:
            raise ObjectParseError(f"需要的键不存在: {e}")

    @property
    def tracks(self) -> list[Self]:
        return [self]

    def link(self) -> str:
        return f"https://music.163.com/song?id={self.id}"

    def prettyString(self) -> str:
        return f"{"/".join(self.artists)} - {self.name}"


@dataclass
class NCMAlbum:
    id: int
    name: str
    tracks: list[NCMTrack]

    @classmethod
    def fromApi(cls, response: Response) -> Self:
        try:
            data: dict = response.json()
        except JSONDecodeError:
            raise ObjectParseError("无法以预期的 Json 格式解析响应")

        if data.get("code") != 200:
            raise ObjectParseError(f"响应码不为 200: {data["code"]}")

        album = data.get("album")
        if album is None:
            raise ObjectParseError("不存在专辑对应的结构")

        try:
            return cls(
                id=album["id"],
                name=album["name"],
                tracks=[NCMTrack.fromData(track) for track in data["songs"]],
            )
        except KeyError as e:
            raise ObjectParseError(f"需要的键不存在: {e}")

    def link(self) -> str:
        return f"https://music.163.com/album?id={self.id}"


@dataclass
class NCMPlaylist:
    id: int
    name: str
    tracks: list[NCMTrack]
    trackIds: list[int]

    @classmethod
    def fromApi(cls, response: Response) -> Self:
        try:
            data: dict = response.json()
        except JSONDecodeError:
            raise ObjectParseError("无法以预期的 Json 格式解析响应")

        if data.get("code") != 200:
            raise ObjectParseError(f"响应码不为 200: {data["code"]}")

        playlist = data.get("playlist")
        if playlist is None:
            raise ObjectParseError("不存在歌单对应的结构")

        try:
            tracks: list[NCMTrack] = []
            trackIds: list[int] = [track["id"] for track in playlist["trackIds"]]

            for track in playlist["tracks"]:
                parsedTrack = NCMTrack.fromData(track)
                trackIds.remove(parsedTrack.id)
                tracks.append(parsedTrack)

            return cls(
                id=playlist["id"],
                name=playlist["name"],
                tracks=tracks,
                trackIds=trackIds,
            )
        except KeyError as e:
            raise ObjectParseError(f"需要的键不存在: {e}")

    def link(self) -> str:
        return f"https://music.163.com/playlist?id={self.id}"

    def fillDetailsOfTracks(self, api) -> None:
        self.tracks.extend(api.getDetailsForTracks(self.trackIds))
        self.trackIds.clear()


@dataclass
class NCMLyrics:
    id: int | None
    isPureMusic: bool
    lyrics: dict[LrcType, str]

    @classmethod
    def fromApi(cls, response: Response) -> Self:
        try:
            data: dict = response.json()
        except JSONDecodeError:
            raise ObjectParseError("无法以预期的 Json 格式解析响应")

        if data.get("code") != 200:
            raise ObjectParseError(f"响应码不为 200: {data["code"]}")

        lyrics: dict[LrcType, str] = {}

        try:
            lyrics[LrcType.Origin] = data["lrc"]["lyric"]
        except KeyError:
            pass

        try:
            lyrics[LrcType.Translation] = data["tlyric"]["lyric"]
        except KeyError:
            pass

        try:
            lyrics[LrcType.Romaji] = data["romalrc"]["lyric"]
        except KeyError:
            pass

        return cls(
            id=None,
            isPureMusic=data.get("pureMusic", False),
            lyrics=lyrics,
        )

    def withId(self, id: int) -> Self:
        self.id = id
        return self

    def get(self, type: LrcType) -> str | None:
        return self.lyrics.get(type, None)



from json import JSONDecodeError
from json import loads as loadJson
from pathlib import Path
from re import Match
from re import compile as compileRegex
from typing import Generator, Iterable, Self

from .constant import CONFIG_LRC_AUTO_MERGE, CONFIG_LRC_AUTO_MERGE_OFFSET
from .enum import LrcMetaType, LrcType
from .error import UnsupportedPureMusicTrackError
from .object import NCMLyrics

__all__ = ["Lrc"]

LRC_RE_COMMIT = compileRegex(r"^\s*#")
LRC_RE_META = compileRegex(r"^\s*\[(?P<type>ti|ar|al|au|length|by|offset):\s*(?P<content>.+?)\s*\]\s*$")
LRC_RE_META_NCM_SPECIAL = compileRegex(r"^\s*\{.*\}\s*$")
LRC_RE_LYRIC = compileRegex(r"^\s*(?P<timeLabels>(?:\s*\[\d{1,2}:\d{1,2}(?:\.\d{1,3})?\])+)\s*(?P<lyric>.+?)\s*$")
LRC_RE_LYRIC_TIMELABEL = compileRegex(r"\[(?P<minutes>\d{1,2}):(?P<seconds>\d{1,2}(?:\.\d{1,3})?)\]")


class Lrc:
    def __init__(self) -> None:
        # metaType: lrcType: metaContent
        self.metadata: dict[LrcMetaType, dict[LrcType, str]] = {}

        # timestamp: lrcType: lrcContent
        self.lyrics: dict[int, dict[LrcType, str]] = {}

        # specials: timestamp/metaType: lrcContent/metaContent
        self.specials: dict[str, list[tuple[int | LrcMetaType, str]]] = {
            "metadata": [],
            "timestamp": [],
        }

    @classmethod
    def fromNCMLyrics(cls, lyrics: NCMLyrics) -> Self:
        if lyrics.isPureMusic:
            raise UnsupportedPureMusicTrackError

        result = cls()

        for lrcType in LrcType:
            lrcStr = lyrics.get(lrcType)
            if lrcStr:
                result.serializeLyricFile(lrcType, lrcStr)

        return result

    def serializeLyricFile(self, lrcType: LrcType, lrcFile: str) -> None:
        self.serializeLyricRows(lrcType, lrcFile.splitlines())

    def serializeLyricRows(self, lrcType: LrcType, lrcRows: Iterable[str]) -> None:
        for row in lrcRows:
            self.serializeLyricRow(lrcType, row)

    def serializeLyricRow(self, lrcType: LrcType, lrcRow: str) -> None:
        # Skip commit lines
        if LRC_RE_COMMIT.match(lrcRow) is not None:
            return

        if LRC_RE_META_NCM_SPECIAL.match(lrcRow) is not None:
            self.appendSpecialNCMMetaDataRow(lrcRow)
            return

        matchedMetaDataRow = LRC_RE_META.match(lrcRow)
        if matchedMetaDataRow is not None:
            self.appendMatchedMetaDataRow(lrcType, matchedMetaDataRow)
            return

        matchedLyricRow = LRC_RE_LYRIC.match(lrcRow)
        if matchedLyricRow is not None:
            self.appendMatchedLyricRow(lrcType, matchedLyricRow)
            return

    def appendLyric(self, lrcType: LrcType, timestamps: Iterable[int], lyric: str):
        for timestamp in timestamps:
            if timestamp in self.lyrics:
                self.lyrics[timestamp][lrcType] = lyric
            else:
                self.lyrics[timestamp] = {lrcType: lyric}

    def appendSpecialNCMMetaDataRow(self, lrcRow: str) -> None:
        try:
            data = loadJson(lrcRow)
        except JSONDecodeError:
            return

        try:
            match len(data["c"]):
                case 1:
                    key, value = data["c"][0]["tx"].replace("：", ":").split(":")
                case 2:
                    key = data["c"][0]["tx"]
                    value = data["c"][1]["tx"]
                case _:
                    return
        except (KeyError, ValueError):
            return

        key = key.strip(" :：")
        value = value.strip()

        self.specials["metadata"].append((LrcMetaType.Author, f"{key}/{value}"))

    def appendMatchedMetaDataRow(self, lrcType: LrcType, matchedLine: Match[str]) -> None:
        metaType, metaContent = matchedLine.groups()

        try:
            metaType = LrcMetaType(metaType)
        except ValueError as e:
            raise ValueError(f"未知的元数据类型：{e}")

        if metaType in self.metadata:
            self.metadata[metaType][lrcType] = metaContent
        else:
            self.metadata[metaType] = {lrcType: metaContent}

    def appendMatchedLyricRow(self, lrcType: LrcType, matchedLine: Match[str]) -> None:
        timeLabels, lyric = matchedLine.groups()
        timestamps: list[int] = []

        for timeLabel in LRC_RE_LYRIC_TIMELABEL.finditer(timeLabels):
            timestamps.append(self._timeLabel2Timestamp(timeLabel))

        if CONFIG_LRC_AUTO_MERGE:
            mergedTimestamps: list[int] = []

            for timestamp in timestamps:
                if timestamp in self.lyrics:
                    mergedTimestamps.append(timestamp)
                else:
                    mergedTimestamps.append(self._mergeOffset(timestamp))

            timestamps = mergedTimestamps

        self.appendLyric(lrcType, timestamps, lyric)

    def deserializeLyricFile(self) -> str:
        return "\n".join(list(self.deserializeLyricRows()))

    def deserializeLyricRows(self) -> Generator[str, None, None]:
        yield from self.generateMetaDataRows()
        yield from self.generateLyricRows()

    def generateMetaDataRows(self) -> Generator[str, None, None]:
        for metaType in LrcMetaType:
            if metaType in self.metadata:
                for lrcType in self.metadata[metaType].keys():
                    yield f"[{metaType.value}:{lrcType.prettyString()}/{self.metadata[metaType][lrcType]}]"

        for metaType, content in self.specials["metadata"]:
            yield f"[{metaType.value}:{content}]"

    def generateLyricRows(self) -> Generator[str, None, None]:
        for timestamp in sorted(self.lyrics.keys()):
            for lrcType in self.lyrics[timestamp].keys():
                yield self._timestamp2TimeLabel(timestamp) + self.lyrics[timestamp][lrcType]

        for timestamp, content in self.specials["timestamp"]:
            yield self._timestamp2TimeLabel(timestamp) + content

    def saveAs(self, path: Path) -> None:
        with path.open("w+") as fs:
            for row in self.deserializeLyricRows():
                fs.write(row)
                fs.write("\n")

    @staticmethod
    def _timeLabel2Timestamp(timeLabel: Match[str]) -> int:
        minutes, seconds = timeLabel.groups()
        return round((int(minutes) * 60 + float(seconds)) * 1000)

    @staticmethod
    def _timestamp2TimeLabel(timestamp: int) -> str:
        seconds = timestamp / 1000
        return f"[{seconds//60:02.0f}:{seconds%60:06.3f}]"

    def _mergeOffset(self, timestamp: int) -> int:
        result = timestamp

        timestampMin = timestamp - CONFIG_LRC_AUTO_MERGE_OFFSET
        timestampMax = timestamp + CONFIG_LRC_AUTO_MERGE_OFFSET

        for existLyric in self.lyrics.keys():
            if timestampMin <= existLyric <= timestampMax:
                result = existLyric
                break

        return result



from http.cookiejar import LoadError, MozillaCookieJar
from json import dumps as dumpJson

from httpx import Client as HttpXClient
from httpx import Request as HttpXRequest
from httpx import Response as HttpXResponse

from .constant import CONFIG_API_DETAIL_TRACK_PER_REQUEST, NCM_API_BASE_URL, PLATFORM
from .error import (
    NCMApiRequestError,
    NCMApiRetryLimitExceededError,
)
from .object import NCMAlbum, NCMLyrics, NCMPlaylist, NCMTrack

try:
    import brotlicffi as brotli  # type: ignore
except ImportError:
    try:
        import brotli
    except ImportError:
        brotli = None

try:
    import zstandard  # type: ignore
except ImportError:
    zstandard = None

try:
    import h2  # type: ignore
except ImportError:
    h2 = None

__all__ = ["NCMApi"]

REQUEST_HEADERS = {
    "Accept": "application/json",
    "Accept-Encoding": f"{"zstd, " if zstandard is not None else ""}{"br, " if brotli is not None else ""}gzip, deflate",
    "Connection": "keep-alive",
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36",
}


class NCMApi:
    def __init__(self) -> None:
        self._cookiePath = PLATFORM.user_config_path / "cookies.txt"
        self._cookieJar = MozillaCookieJar()

        try:
            self._cookieJar.load(str(self._cookiePath))
        except (FileNotFoundError, LoadError):
            pass

        self._httpClient = HttpXClient(
            base_url=NCM_API_BASE_URL,
            cookies=self._cookieJar,
            headers=REQUEST_HEADERS,
            http2=h2 is not None,
        )

    def _fetch(self, request: HttpXRequest, retry: int | None = 4) -> HttpXResponse:
        if retry is not None:  # None => Disable retry
            if retry < 0:
                retry = 0

            while retry >= 0:
                try:
                    return self._httpClient.send(request)
                except Exception:
                    retry -= 1

            raise NCMApiRetryLimitExceededError

        else:
            try:
                return self._httpClient.send(request)
            except Exception:
                raise NCMApiRequestError

    def saveCookies(self) -> None:
        self._cookieJar.save(str(self._cookiePath))

    def getDetailsForTrack(self, trackId: int) -> NCMTrack:
        request = self._httpClient.build_request("GET", "/v3/song/detail", params={"c": f"[{{'id':{trackId}}}]"})
        return NCMTrack.fromApi(self._fetch(request)).pop()

    def getDetailsForTracks(self, trackIds: list[int]) -> list[NCMTrack]:
        result: list[NCMTrack] = []
        seek = 0

        while True:
            seekedTrackIds = trackIds[seek : seek + CONFIG_API_DETAIL_TRACK_PER_REQUEST]

            if len(seekedTrackIds) == 0:
                break

            params = {
                "c": dumpJson(
                    [{"id": trackId} for trackId in seekedTrackIds],
                    separators=(",", ":"),
                )
            }

            request = self._httpClient.build_request("GET", "/v3/song/detail", params=params)

            result.extend(NCMTrack.fromApi(self._fetch(request)))

            seek += CONFIG_API_DETAIL_TRACK_PER_REQUEST

        return result

    def getDetailsForAlbum(self, albumId: int) -> NCMAlbum:
        request = self._httpClient.build_request("GET", f"/v1/album/{albumId}")
        return NCMAlbum.fromApi(self._fetch(request))

    def getDetailsForPlaylist(self, playlistId: int) -> NCMPlaylist:
        request = self._httpClient.build_request("GET", "/v6/playlist/detail", params={"id": playlistId})
        return NCMPlaylist.fromApi(self._fetch(request))

    def getLyricsByTrack(self, trackId: int) -> NCMLyrics:
        params = {
            "id": trackId,
            "cp": False,
            "lv": 0,
            "tv": 0,
            "rv": 0,
            "kv": 0,
            "yv": 0,
            "ytv": 0,
            "yrv": 0,
        }

        request = self._httpClient.build_request("GET", "/song/lyric/v1", params=params)
        return NCMLyrics.fromApi(self._fetch(request)).withId(trackId)



from httpx import RequestError

__all__ = [
    "NCMLyricsAppError",
    "NCMApiError",
    "NCMApiRequestError",
    "NCMApiRetryLimitExceededError",
    "ObjectParseError",
    "ParseLinkError",
    "UnsupportedLinkError",
    "UnsupportedPureMusicTrackError",
]


class NCMLyricsAppError(Exception):
    """NCMLyrics 错误"""


class NCMApiError(NCMLyricsAppError):
    """使用网易云音乐 API 时出现错误"""


class NCMApiRequestError(NCMApiError, RequestError):
    """请求网易云音乐 API 时出现错误"""


class NCMApiRetryLimitExceededError(NCMApiError):
    """请求网易云音乐 API 时错误次数超过重试次数上限"""


class ObjectParseError(NCMLyricsAppError):
    """解析网易云音乐 API 返回的数据时出现错误"""


class ParseLinkError(NCMLyricsAppError):
    """无法解析此分享链接"""


class UnsupportedLinkError(NCMLyricsAppError):
    """不支持的分享链接"""


class UnsupportedPureMusicTrackError(NCMLyricsAppError):
    """不支持纯音乐单曲"""
